# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT

__doc__ = """
  Allows to run the foss_cli with: python -m fossology
"""
from fossology.foss_cli import main

if __name__ == "__main__":
    main()
